
import logging
from typing import Any, Optional

from brownie import chain
from multicall.utils import await_awaitable
from y import convert
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)

#yLazyLogger(logger)
def is_froyo(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return chain.id == Network.Fantom and address == '0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3'

#yLazyLogger(logger)
def get_price(token: Any, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_price_async(token, block))
    

async def get_price_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    pool = '0x83E5f18Da720119fF363cF63417628eB0e9fd523'
    virtual_price = await raw_call_async(pool, "get_virtual_price()", block=block, output='int')
    return UsdPrice(virtual_price / 1e18)



import asyncio
import logging
from functools import cached_property, lru_cache
from typing import Any, List, Optional, Union

from async_lru import alru_cache
from brownie import chain
from multicall import Call
from multicall.utils import await_awaitable
from y import convert
from y.classes.common import ERC20, ContractBase
from y.classes.singleton import Singleton
from y.contracts import Contract
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice
from y.exceptions import ContractNotVerified
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.multicall import fetch_multicall
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)


v1_pools = {
    Network.Mainnet: [
        "0x398eC7346DcD622eDc5ae82352F02bE94C62d119"
    ],
}.get(chain.id, [])

v2_pools = {
    Network.Mainnet: [
        "0x7d2768dE32b0b80b7a3454c06BdAc94A69DDc7A9", # aave
        "0x7937D4799803FbBe595ed57278Bc4cA21f3bFfCB", # aave amm
        "0xcE744a9BAf573167B2CF138114BA32ed7De274Fa", # umee
    ],
    Network.Polygon: [
        "0x8dFf5E27EA6b7AC08EbFdf9eB090F32ee9a30fcf", # aave
    ],
    Network.Avalanche: [
        "0x70BbE4A294878a14CB3CDD9315f5EB490e346163", # blizz
    ]
}.get(chain.id, [])        


class AaveMarketBase(ContractBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __contains__(self, __o: object) -> bool:
        return convert.to_address(__o) in self.atokens


class AaveMarketV1(AaveMarketBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __repr__(self) -> str:
        return f"<AaveMarketV1 '{self.address}'>"
    
    #yLazyLogger(logger)
    @lru_cache
    def underlying(self, token_address: AddressOrContract) -> ERC20:
        return await_awaitable(self.underlying_async(token_address))
    
    #yLazyLogger(logger)
    @alru_cache
    async def underlying_async(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call_async(token_address, 'underlyingAssetAddress()',output='address')
        return ERC20(underlying)
    
    @cached_property
    #yLazyLogger(logger)
    def atokens(self) -> List[ERC20]:
        reserves_data = Call(self.address, ['getReserves()(address[])'], [[self.address,None]])()[self.address]
        reserves_data = fetch_multicall(*[[self.contract, 'getReserveData', reserve] for reserve in reserves_data])
        atokens = [ERC20(reserve['aTokenAddress']) for reserve in reserves_data]
        logger.info(f'loaded {len(atokens)} v1 atokens for {self.__repr__()}')
        return atokens


class AaveMarketV2(AaveMarketBase):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def __repr__(self) -> str:
        return f"<AaveMarketV2 '{self.address}'>"
    
    #yLazyLogger(logger)
    @lru_cache
    def underlying(self, token_address: AddressOrContract) -> ERC20:
        return await_awaitable(self.underlying_async(token_address))
    
    #yLazyLogger(logger)
    @alru_cache
    async def underlying_async(self, token_address: AddressOrContract) -> ERC20:
        underlying = await raw_call_async(token_address, 'UNDERLYING_ASSET_ADDRESS()',output='address')
        print(underlying)
        return ERC20(underlying)

    @cached_property
    #yLazyLogger(logger)
    def atokens(self) -> List[ERC20]:
        reserves = Call(self.address, ['getReservesList()(address[])'], [[self.address,None]])()[self.address]
        reserves_data = [
            Call(
                self.address,
                ['getReserveData(address)((uint256,uint128,uint128,uint128,uint128,uint128,uint40,address,address,address,address,uint8))',reserve],
            )()
            for reserve in reserves
        ]

        try:
            atokens = [ERC20(reserve_data[7]) for reserve_data in reserves_data]
            logger.info(f'loaded {len(atokens)} v2 atokens for {self.__repr__()}')
            return atokens
        except TypeError: # TODO figure out what to do about non verified aave markets
            logger.error(f'failed to load tokens for {self.__repr__()}')
            return []


class AaveRegistry(metaclass = Singleton):
    def __init__(self) -> None:
        pass

    @cached_property
    def pools(self) -> List[Union[AaveMarketV1, AaveMarketV2]]:
        return self.pools_v1 + self.pools_v2
    
    @cached_property
    def pools_v1(self) -> List[AaveMarketV1]:
        return [AaveMarketV1(pool) for pool in v1_pools]
    
    @cached_property
    def pools_v2(self) -> List[AaveMarketV2]:
        return [AaveMarketV2(pool) for pool in v2_pools]
    
    #yLazyLogger(logger)
    def pool_for_atoken(self, token_address: AnyAddressType) -> Optional[Union[AaveMarketV1, AaveMarketV2]]:
        for pool in self.pools:
            if token_address in pool:
                return pool

    def __contains__(self, __o: object) -> bool:
        return any(__o in pool for pool in self.pools)

    #yLazyLogger(logger)
    @lru_cache
    def is_atoken(self, token_address: AnyAddressType) -> bool:
        return token_address in self
    
    async def is_wrapped_atoken(self, token_address: AnyAddressType) -> bool:
        try:
            contract = await Contract.coroutine(token_address)
            attrs = "ATOKEN", "STATIC_ATOKEN_LM_REVISION", "staticToDynamicAmount"
            return all(hasattr(contract, attr) for attr in attrs)
        except ContractNotVerified:
            return False
    
    #yLazyLogger(logger)
    @lru_cache
    def underlying(self, token_address: AddressOrContract) -> ERC20:
        return await_awaitable(self.underlying_async(token_address))
    
    #yLazyLogger(logger)
    @alru_cache
    async def underlying_async(self, token_address: AddressOrContract) -> ERC20:
        pool = self.pool_for_atoken(token_address)
        return await pool.underlying_async(token_address)
    
    #yLazyLogger(logger)
    def get_price(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_price_async(token_address, block=block))
    
    #yLazyLogger(logger)
    async def get_price_async(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        underlying = await self.underlying_async(token_address)
        return await underlying.price_async(block)
    
    async def get_price_wrapped_async(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        contract, scale = await asyncio.gather(Contract.coroutine(token_address), ERC20(token_address).scale)
        underlying, price_per_share = await asyncio.gather(
            contract.ATOKEN.coroutine(block_identifier=block),
            contract.staticToDynamicAmount.coroutine(scale, block_identifier=block),
        )
        price_per_share /= scale
        return price_per_share * await ERC20(underlying).price_async(block)


aave = AaveRegistry()




import asyncio
import logging
from functools import lru_cache
from typing import Any, List, Optional

from async_lru import alru_cache
from multicall import Call
from multicall.utils import await_awaitable
from y.classes.common import ERC20, WeiBalance
from y.contracts import has_methods_async
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

#yLazyLogger(logger)
@lru_cache(maxsize=None)
def is_token_set(token: AnyAddressType) -> bool:
    return await_awaitable(is_token_set_async(token))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_token_set_async(token: AnyAddressType) -> bool:
    return any(await asyncio.gather(
        has_methods_async(token, ("getComponents()(address[])", "naturalUnit()(uint)")),
        has_methods_async(token, ("getComponents()(address[])", "getModules()(address[])", "getPositions()(address[])")),
    ))

#yLazyLogger(logger)
def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return TokenSet(token).get_price(block=block)

#yLazyLogger(logger)
async def get_price_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await TokenSet(token).get_price_async(block=block)


class TokenSet(ERC20):
    def __init__(self, address: AnyAddressType, *args: Any, **kwargs: Any) -> None:
        super().__init__(address, *args, **kwargs)
    
    def components(self, block: Optional[Block] = None) -> List[ERC20]:
        return await_awaitable(self.components_async(block=block))

    @alru_cache(maxsize=100)
    async def components_async(self, block: Optional[Block] = None) -> List[ERC20]:
        components = await self.contract.getComponents.coroutine(block_identifier = block)
        return [ERC20(component) for component in components]
    
    def balances(self, block: Optional[Block] = None) -> List[WeiBalance]:
        return await_awaitable(self.balances_async(block=block))
    
    async def balances_async(self, block: Optional[Block] = None) -> List[WeiBalance]:
        if hasattr(self.contract, 'getUnits'):
            balances = await self.contract.getUnits.coroutine(block_identifier = block)
        elif hasattr(self.contract, 'getTotalComponentRealUnits'):
            balances = await asyncio.gather(*[
                Call(self.address, ["getTotalComponentRealUnits(address)(uint)", component.address], block_id=block).coroutine()
                for component in await self.components_async(block)
            ])
        return [WeiBalance(balance, component, block=block) for component, balance in zip(await self.components_async(block=block), balances)]

    def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_price_async(block=block))
    
    async def get_price_async(self, block: Optional[Block] = None) -> UsdPrice:
        total_supply = await self.total_supply_readable_async(block=block)
        if total_supply == 0:
            return UsdPrice(0)
        return UsdPrice(sum(await asyncio.gather(*[balance.value_usd_async for balance in await self.balances_async(block=block)])))




import _thread
import functools
import logging
import os
from typing import Any, Callable

DEBUG = os.environ.get('YPRICEMAGIC_DEBUG', False)

def continue_on_revert(func: Callable) -> Any:
    '''
    Decorates a call-making function. If the call reverts, returns `None`. Raises any other exceptions.
    '''
    from y.exceptions import continue_if_call_reverted

    @functools.wraps(func)
    def continue_on_revert_wrap(*args: Any, **kwargs: Any) -> Callable:
        try:
            return func(*args,**kwargs)
        except Exception as e:
            continue_if_call_reverted(e)
    
    return continue_on_revert_wrap

def wait_or_exit_after(func):
    @functools.wraps(func)
    def wrap(self):
        func(self)
        self._done.wait()
        if self._has_exception:
            logging.error(self._exception)
            _thread.interrupt_main()
    return wrap



import asyncio
import logging
from typing import Optional

from brownie import chain
from multicall.utils import await_awaitable
from y import convert
from y.constants import weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices import magic
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)

class wstEth:
    def __init__(self) -> None:
        try:
            self.address = {
                Network.Mainnet: '0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0'
            }[chain.id]
        except KeyError:
            self.address = None
    
    #yLazyLogger(logger)
    def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_price_async(block=block))

    #yLazyLogger(logger)
    async def get_price_async(self, block: Optional[Block] = None) -> UsdPrice:
        share_price, weth_price = await asyncio.gather(
            raw_call_async(self.address, 'stEthPerToken()', output='int', block=block),
            magic.get_price_async(weth, block),
        )
        share_price /= 1e18
        return UsdPrice(share_price * weth_price)

wsteth = wstEth()

#yLazyLogger(logger)
def is_wsteth(address: AnyAddressType) -> bool:
    if chain.id != Network.Mainnet:
        return False
    address = convert.to_address(address)
    return address == wsteth.address





import asyncio
import logging
from typing import Optional

from async_lru import alru_cache
from multicall.utils import await_awaitable
from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, has_methods_async
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)


#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_mstable_feeder_pool(address: AnyAddressType) -> bool:
    return await has_methods_async(address, ('getPrice()((uint,uint))','mAsset()(address)'))

async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_price_async(token, block))

#yLazyLogger(logger)
async def get_price_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    contract = await Contract.coroutine(address)
    ratio, masset, scale = await asyncio.gather(
        contract.getPrice.coroutine(block_identifier=block),
        contract.mAsset.coroutine(block_identifier=block),
        ERC20(address).scale,
    )
    ratio = ratio[0] / scale
    underlying_price = await magic.get_price_async(masset, block)
    return UsdPrice(underlying_price * ratio)



import asyncio
import logging
from typing import Any, Callable, KeysView, List, Optional, Tuple, Union

import brownie
from brownie.convert.datatypes import EthAddress
from multicall.utils import await_awaitable

from y.contracts import Contract
from y.datatypes import Address, AddressOrContract, Block
from y.utils.multicall import (multicall_decimals_async,
                               multicall_totalSupply_async)
from y.utils.raw_calls import _decimals, _totalSupply

logger = logging.getLogger(__name__)

SUPPORTED_INPUT_TYPES = str, Address, EthAddress, brownie.Contract, Contract

# These helpers can be used to fetch values for one or more tokens at once.

def decimals(
    contract_address_or_addresses: Union[AddressOrContract,List[AddressOrContract],Tuple[AddressOrContract]],
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ): 
    return await_awaitable(
        decimals_async(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)
    )


async def decimals_async(
    contract_address_or_addresses: Union[AddressOrContract,List[AddressOrContract],Tuple[AddressOrContract]],
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ): 

    func = _choose_appropriate_fn(contract_address_or_addresses, _decimals, multicall_decimals_async)
    return await func(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)


def totalSupply(
    contract_address_or_addresses: Union[AddressOrContract,List[AddressOrContract],Tuple[AddressOrContract]],
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ):
    return await_awaitable(
        totalSupply_async(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)
    )


async def totalSupply_async(
    contract_address_or_addresses: Union[AddressOrContract,List[AddressOrContract],Tuple[AddressOrContract]],
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ):

    func = _choose_appropriate_fn(contract_address_or_addresses, _totalSupply, multicall_totalSupply_async)
    return await func(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)


def totalSupplyReadable(
    contract_address_or_addresses: Union[AddressOrContract,List[AddressOrContract],Tuple[AddressOrContract]],
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ):
    return await_awaitable(
        totalSupplyReadable_async(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)
    )


async def totalSupplyReadable_async(
    contract_address_or_addresses: Union[AddressOrContract,List[AddressOrContract],Tuple[AddressOrContract]],
    block: Optional[Block] = None, 
    return_None_on_failure: bool = False
    ):

    token_supplys, token_decimals = await asyncio.gather(
        totalSupply_async(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure),
        decimals_async(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure),
    )

    if type(token_supplys) == brownie.Wei: # if only fetching totalSupply for one token
        supply = token_supplys
        decimal = token_decimals
        return supply / 10 ** decimal
    return [supply / 10 ** decimal for supply, decimal in zip(token_supplys, token_decimals)]


def _choose_appropriate_fn(
    input: Any,
    singlecall_fn: Callable,
    multicall_fn: Callable
    ):

    input_type = _input_type(input)
    if input_type == 'single': return singlecall_fn
    elif input_type == 'multi': return multicall_fn


def _input_type(
    input: Any
    ) -> str:
    
    if type(input) in [list, tuple] or isinstance(input, KeysView): 
        for value in input: _check_if_supported(value)
        return 'multi'
    
    else:
        _check_if_supported(input)
        return 'single'


def _check_if_supported(
    input: Any
    ) -> bool:
    
    if type(input) not in SUPPORTED_INPUT_TYPES:
        raise TypeError(f'Unsupported input type: {type(input)}')



import asyncio
import logging
from typing import Optional

from brownie import chain
from multicall.utils import await_awaitable
from y import convert
from y.classes.common import ERC20
from y.constants import weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices import magic
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call_async

logger = logging.getLogger(__name__)

#yLazyLogger(logger)
def is_creth(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return chain.id == Network.Mainnet and address == '0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd'

#yLazyLogger(logger)
def get_price_creth(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_price_creth_async(token, block=block))

#yLazyLogger(logger)
async def get_price_creth_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    total_balance, total_supply, weth_price = await asyncio.gather(
        raw_call_async(address, 'accumulated()', output='int', block=block),
        ERC20(address).total_supply_async(block),
        magic.get_price_async(weth,block),
    )
    per_share = total_balance / total_supply
    return per_share * weth_price



import asyncio
import logging
from functools import cached_property
from typing import List, Optional

from async_lru import alru_cache
from async_property import async_cached_property
from brownie import chain
from brownie.convert.datatypes import EthAddress, HexString
from eth_abi import encode_single
from multicall import Call
from multicall.utils import await_awaitable

from y import convert
from y.classes.singleton import Singleton
from y.contracts import Contract, has_method_async
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.exceptions import UnsupportedNetwork, call_reverted
from y.networks import Network

logger = logging.getLogger(__name__)

addresses = {
    Network.Mainnet: '0x823bE81bbF96BEc0e25CA13170F5AaCb5B79ba83',
}


class Synthetix(metaclass=Singleton):
    def __init__(self) -> None:
        if chain.id not in addresses:
            raise UnsupportedNetwork("synthetix is not supported on this network")
    
    @async_cached_property
    async def address_resolver(self) -> Contract:
        return await Contract.coroutine(addresses[chain.id])

    @alru_cache(maxsize=256)
    async def get_address(self, name: str, block: Block = None) -> Contract:
        """
        Get contract from Synthetix registry.
        See also https://docs.synthetix.io/addresses/
        """
        address_resolver = await self.address_resolver
        address = await address_resolver.getAddress.coroutine(encode_single('bytes32', name.encode()), block_identifier=block)
        proxy = await Contract.coroutine(address)
        return await Contract.coroutine(proxy.target()) if hasattr(proxy, 'target') else proxy

    @cached_property
    async def synths(self) -> List[EthAddress]:
        """
        Get target addresses of all synths.
        """
        proxy_erc20 = await self.get_address('ProxyERC20')
        synth_count = await proxy_erc20.availableSynthCount.coroutine()
        synths = await asyncio.gather(*[proxy_erc20.availableSynths(i) for i in range(synth_count)])
        logger.info(f'loaded {len(synths)} synths')
        return synths
    '''
    def __contains__(self, token: AnyAddressType) -> bool:
        """
        Check if a token is a synth.
        """
        token = convert.to_address(token)
        if synthetix.get_currency_key(token):
            return True
        if has_method(token, 'target()(address)'):
            target = Call(token, 'target()(address)')()
            return target in synthetix.synths and Call(target, 'proxy()(address)')() == token
        return False
    '''
    async def is_synth(self, token: AnyAddressType) -> bool:
        """
        Check if a token is a synth.
        """
        token = convert.to_address(token)
        if await synthetix.get_currency_key_async(token):
            return True
        if await has_method_async(token, 'target()(address)'):
            target = await Call(token, 'target()(address)').coroutine()
            return target in await synthetix.synths and await Call(target, 'proxy()(address)').coroutine() == token
        return False

    def get_currency_key(self, token: AnyAddressType) -> Optional[HexString]:
        return await_awaitable(self.get_currency_key_async(token))
    
    @alru_cache(maxsize=None)
    async def get_currency_key_async(self, token: AnyAddressType) -> Optional[HexString]:
        target = await Call(token, ['target()(address)']).coroutine() if await has_method_async(token,'target()(address)') else token
        return await Call(target, ['currencyKey()(bytes32)']).coroutine() if await has_method_async(token,'currencyKey()(bytes32)') else None

    def get_price(self, token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        return await_awaitable(self.get_price_async(token, block=block))
    
    async def get_price_async(self, token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        """
        Get a price of a synth in dollars.
        """
        token = convert.to_address(token)
        try:
            rates, key = await asyncio.gather(self.get_address('ExchangeRates', block=block), self.get_currency_key_async(token))
            return UsdPrice(await rates.rateForCurrency.coroutine(key, block_identifier=block) / 10 ** 18)
        except Exception as e:
            if call_reverted(e):
                return None
            raise


try:
    synthetix = Synthetix()
except UnsupportedNetwork:
    synthetix = set()



import asyncio
from functools import lru_cache
from typing import Optional, Tuple

from async_lru import alru_cache
from brownie.exceptions import ContractNotFound
from multicall import Call
from multicall.utils import await_awaitable
from y import Contract
from y.classes.common import ERC20, WeiBalance
from y.datatypes import AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import ContractNotVerified, MessedUpBrownieContract
from y.utils.cache import memory


@memory.cache()
def is_generic_amm(lp_token_address: AnyAddressType) -> bool:
    try:
        token_contract = Contract(lp_token_address)
        return all(hasattr(token_contract, attr) for attr in ['getReserves','token0','token1'])
    except (ContractNotFound, ContractNotVerified):
        return False
    except MessedUpBrownieContract:
        # probably false, can get more specific when there's a need.
        return False
        
class GenericAmm:
    def __contains__(self, lp_token_address: AnyAddressType) -> bool:
        return is_generic_amm(lp_token_address)
    
    def get_price(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_price_async(lp_token_address, block=block))
    
    async def get_price_async(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        tvl, total_supply = await asyncio.gather(
            self.get_tvl_async(lp_token_address, block=block),
            ERC20(lp_token_address).total_supply_readable_async(block=block),
        )
        if total_supply is None:
            return None
        elif total_supply == 0:
            return 0
        return UsdPrice(tvl / total_supply)

    @lru_cache(maxsize=None)
    def get_tokens(self, lp_token_address: AnyAddressType) -> Tuple[ERC20,ERC20]:
        return await_awaitable(self.get_tokens_async(lp_token_address))
    
    @alru_cache(maxsize=None)
    async def get_tokens_async(self, lp_token_address: AnyAddressType) -> Tuple[ERC20,ERC20]:
        tokens = await asyncio.gather(
            Call(lp_token_address, ['token0()(address)']).coroutine(),
            Call(lp_token_address, ['token1()(address)']).coroutine(),
        )
        return ERC20(tokens[0]), ERC20(tokens[1])
    
    def get_tvl(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdValue:
        return await_awaitable(self.get_tvl_async(lp_token_address, block=block))
    
    async def get_tvl_async(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdValue:
        lp_token_contract = await Contract.coroutine(lp_token_address)
        tokens = await self.get_tokens_async(lp_token_address)
        reserves = await lp_token_contract.getReserves.coroutine(block_identifier=block)
        reserves = [WeiBalance(reserve,token,block) for token, reserve in zip(tokens,reserves)]
        return UsdValue(sum(await asyncio.gather(*[reserve.value_usd_async for reserve in reserves])))


generic_amm = GenericAmm()




import asyncio
import logging
from typing import Optional

from async_lru import alru_cache
from multicall.utils import await_awaitable
from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, has_methods_async
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils.cache import memory
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

@memory.cache()
#yLazyLogger(logger)
def is_ib_token(token: AnyAddressType) -> bool:
    return await_awaitable(is_ib_token_async(token))

#yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_ib_token_async(token: AnyAddressType) -> bool:
    return await has_methods_async(token, ('debtShareToVal(uint)(uint)','debtValToShare(uint)(uint)','token()(address)','totalToken()(uint)','totalSupply()(uint)'))

#yLazyLogger(logger)
def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    return await_awaitable(get_price_async(token, block=block))

#yLazyLogger(logger)
async def get_price_async(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    contract = await Contract.coroutine(address)
    token, total_bal, total_supply = await asyncio.gather(
        contract.token.coroutine(block_identifier=block),
        contract.totalToken.coroutine(block_identifier=block),
        contract.totalSupply.coroutine(block_identifier=block),
    )
    token_scale, pool_scale = await asyncio.gather(ERC20(token).scale, ERC20(address).scale)
    total_bal /= token_scale
    total_supply /= pool_scale
    share_price = total_bal / total_supply
    token_price = await magic.get_price_async(token, block)
    price = share_price * token_price
    return price



import datetime
import logging
from typing import Union

from async_lru import alru_cache
from brownie import chain, web3

from y.exceptions import NoBlockFound
from y.networks import Network
from y.utils.cache import memory
from y.utils.client import get_ethereum_client, get_ethereum_client_async
from y.utils.dank_mids import dank_w3
from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)

@memory.cache()
@yLazyLogger(logger)
def get_block_timestamp(height: int) -> int:
    client = get_ethereum_client()
    if client in ['tg', 'erigon']:
        header = web3.manager.request_blocking(f"{client}_getHeaderByNumber", [height])
        return int(header.timestamp, 16)
    else:
        return chain[height].timestamp

@alru_cache(maxsize=None, cache_exceptions=False)
async def get_block_timestamp_async(height: int) -> int:
    client = await get_ethereum_client_async()
    if client in ['tg', 'erigon']:
        header = await dank_w3.manager.coro_request(f"{client}_getHeaderByNumber", [height])
        return int(header.timestamp, 16)
    else:
        block = await dank_w3.eth.get_block(height)
        return block.timestamp

@memory.cache()
#yLazyLogger(logger)
def last_block_on_date(date: Union[str, datetime.date]) -> int:
    """ Returns the last block on a given `date`. You can pass either a `datetime.date` object or a date string formatted as 'YYYY-MM-DD'. """
    if isinstance(date, datetime.datetime):
        raise TypeError(
            "You can not pass in a `datetime.datetime` object. Please call date() on your input before passing it to this funciton."
        )
    if not isinstance(date, datetime.date):
        date = datetime.datetime.strptime(date, "%Y-%m-%d").date()
    
    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        logger.debug('block: ' + str(mid))
        mid_ts = get_block_timestamp(mid)
        mid_date = datetime.date.fromtimestamp(mid_ts)
        logger.debug(f'mid: {mid_date}')
        logger.debug(date)
        if mid_date > date:
            hi = mid
        else:
            lo = mid
    hi = hi - 1
    block = hi if hi != height else None
    logger.debug(f'last {Network.name()} block on date {date} -> {block}')
    return block


#yLazyLogger(logger)
def closest_block_after_timestamp(timestamp: int) -> int:
    try:
        block = _closest_block_after_timestamp_cached(timestamp)
    except NoBlockFound:
        block = None
    logger.debug(f'closest {Network.name()} block after timestamp {timestamp} -> {block}')
    return block

@alru_cache(maxsize=1024, cache_exceptions=False)
async def closest_block_after_timestamp_async(timestamp: int) -> int:
    height = await dank_w3.eth.block_number
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        if await get_block_timestamp_async(mid) > timestamp:
            hi = mid
        else:
            lo = mid
    if hi == height:
        raise NoBlockFound(f"No block found after timestamp {timestamp}")
    return hi


@memory.cache()
def _closest_block_after_timestamp_cached(timestamp: int) -> int:
    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        if get_block_timestamp(mid) > timestamp:
            hi = mid
        else:
            lo = mid
    if hi == height:
        raise NoBlockFound(f"No block found after timestamp {timestamp}")
    return hi



import asyncio
import logging
from typing import Dict, List, Optional, Tuple

from brownie import chain
from brownie.convert.datatypes import EthAddress
from brownie.exceptions import VirtualMachineError
from multicall.utils import await_awaitable

from y.classes.common import ERC20
from y.classes.singleton import Singleton
from y.constants import dai, usdc, wbtc, weth
from y.contracts import (Contract, contract_creation_block_async,
                         has_methods_async)
from y.datatypes import (AddressOrContract, AnyAddressType, Block, UsdPrice,
                         UsdValue)
from y.networks import Network
from y.prices import magic
from y.utils.logging import yLazyLogger
from y.utils.multicall import fetch_multicall, multicall_decimals_async

EXCHANGE_PROXY = {
    Network.Mainnet: '0x3E66B66Fd1d0b02fDa6C811Da9E0547970DB2f21',
}.get(chain.id, None)

logger = logging.getLogger(__name__)

async def _calc_out_value(token_out: AddressOrContract, total_outout: int, scale: float, block) -> float:
    out_scale, out_price = await asyncio.gather(ERC20(token_out).scale, magic.get_price_async(token_out, block))
    return (total_outout / out_scale) * out_price / scale

class BalancerV1Pool(ERC20):
    def __init__(self, pool_address: AnyAddressType) -> None:
        super().__init__(pool_address)

    #yLazyLogger(logger)
    def tokens(self, block: Optional[Block] = None) -> List[ERC20]:
        tokens = self.contract.getCurrentTokens(block_identifier=block)
        return [ERC20(token) for token in tokens]

    #yLazyLogger(logger)
    def get_pool_price(self, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_pool_price_async(block=block))

    #yLazyLogger(logger)
    async def get_pool_price_async(self, block: Optional[Block] = None) -> UsdPrice:
        supply = await self.total_supply_readable_async(block=block)
        if supply == 0:
            return 0
        return UsdPrice(await self.get_tvl_async(block=block) / supply)

    #yLazyLogger(logger)
    def get_tvl(self, block: Optional[Block] = None) -> Optional[UsdValue]:
        return await_awaitable(self.get_tvl_async(block=block))

    #yLazyLogger(logger)
    async def get_tvl_async(self, block: Optional[Block] = None) -> Optional[UsdValue]:
        token_balances = await self.get_balances_async()
        good_balances = {
            token: balance
            for token, balance
            in token_balances.items()
            if await token.price_async(block=block, return_None_on_failure=True) is not None
        }
        
        prices = await asyncio.gather(*[token.price_async(block=block, return_None_on_failure = True) for token in good_balances])

        # in case we couldn't get prices for all tokens, we can extrapolate from the prices we did get
        good_value = sum(balance * price for balance, price in zip(good_balances.values(),prices))
        if len(good_balances):
            return good_value / len(good_balances) * len(token_balances)
        return None

    #yLazyLogger(logger)
    def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, float]:
        return await_awaitable(self.get_balances_async(block=block))
    
    #yLazyLogger(logger)
    async def get_balances_async(self, block: Optional[Block] = None) -> Dict[ERC20, float]:
        tokens = self.tokens(block=block)
        balances = fetch_multicall(*[[self.contract, "getBalance", token] for token in tokens], block=block)
        balances = [balance if balance else 0 for balance in balances]
        decimals = await multicall_decimals_async(tokens, block)
        return {token:balance / 10 ** decimal for token, balance, decimal in zip(tokens,balances,decimals)}


class BalancerV1(metaclass=Singleton):
    def __init__(self) -> None:
        self.exchange_proxy = Contract(EXCHANGE_PROXY) if EXCHANGE_PROXY else None
    
    def __str__(self) -> str:
        return "BalancerV1()"
    
    def __repr__(self) -> str:
        return "<BalancerV1>"
    
    #yLazyLogger(logger)
    def is_pool(self, token_address: AnyAddressType) -> bool:
        return await_awaitable(self.is_pool_async(token_address))
    
    #yLazyLogger(logger)
    async def is_pool_async(self, token_address: AnyAddressType) -> bool:
        return await has_methods_async(token_address ,("getCurrentTokens()(address[])", "getTotalDenormalizedWeight()(uint)", "totalSupply()(uint)"))
    
    #yLazyLogger(logger)
    def get_pool_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        return await_awaitable(self.get_pool_price_async(token_address, block=block))

    #yLazyLogger(logger)
    async def get_pool_price_async(self, token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        assert await self.is_pool_async(token_address)
        return await BalancerV1Pool(token_address).get_pool_price_async(block=block)

    #yLazyLogger(logger)
    def get_token_price(self, token_address: AddressOrContract, block: Optional[Block] = None) -> Optional[UsdPrice]:
        return await_awaitable(self.get_token_price_async(token_address, block=block))
    
    #yLazyLogger(logger)
    async def get_token_price_async(self, token_address: AddressOrContract, block: Optional[Block] = None) -> Optional[UsdPrice]:
        if block is not None and block < await contract_creation_block_async(self.exchange_proxy, True):
            return None
        scale = 1.0
        out, totalOutput = await self.get_some_output_async(token_address, block=block)
        if out:
            return await _calc_out_value(out, totalOutput, scale, block=block)
        # Can we get an output if we try smaller size?
        scale = 0.5
        out, totalOutput = await self.get_some_output_async(token_address, block=block, scale=scale) 
        if out:
            return await _calc_out_value(out, totalOutput, scale, block=block)
        # How about now? 
        scale = 0.1
        out, totalOutput = await self.get_some_output_async(token_address, block=block, scale=scale)
        if out:
            return await _calc_out_value(out, totalOutput, scale, block=block)
        return None

    #yLazyLogger(logger)
    def check_liquidity_against(
        self,
        token_in: AddressOrContract,
        token_out: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Tuple[EthAddress, int]:

        return await_awaitable(
            self.check_liquidity_against_async(token_in, token_out, scale=scale, block=block)
        )
    
    #yLazyLogger(logger)
    async def check_liquidity_against_async(
        self,
        token_in: AddressOrContract,
        token_out: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Tuple[EthAddress, int]:

        amount_in = await ERC20(token_in).scale * scale
        view_split_exact_in = await self.exchange_proxy.viewSplitExactIn.coroutine(
            token_in,
            token_out,
            amount_in,
            32, # NOTE: 32 is max
            block_identifier = block,
        )

        output = view_split_exact_in['totalOutput']

        return token_out, output

    #yLazyLogger(logger)
    def get_some_output(
        self,
        token_in: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Tuple[EthAddress,int]:

        return await_awaitable(self.get_some_output_async(token_in, scale=scale, block=block))
    
    #yLazyLogger(logger)
    async def get_some_output_async(
        self,
        token_in: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Tuple[EthAddress,int]:

        try:
            out, totalOutput = await self.check_liquidity_against_async(token_in, weth, block=block)
        except (ValueError, VirtualMachineError):
            try:
                out, totalOutput = await self.check_liquidity_against_async(token_in, dai, block=block)
            except (ValueError, VirtualMachineError):
                try:
                    out, totalOutput = await self.check_liquidity_against_async(token_in, usdc, block=block)
                except (ValueError, VirtualMachineError):
                    try:
                        out, totalOutput = await self.check_liquidity_against_async(token_in, wbtc, block=block)
                    except (ValueError, VirtualMachineError):
                        out = None
                        totalOutput = None
        return out, totalOutput



import asyncio
import logging
from functools import cached_property, lru_cache
from typing import List, Optional, Union

from async_lru import alru_cache
from brownie import chain
from multicall.utils import await_awaitable
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices.dex.balancer.v1 import BalancerV1
from y.prices.dex.balancer.v2 import BalancerV2

logger = logging.getLogger(__name__)


class BalancerMultiplexer:
    def __init__(self) -> None:
        pass
    
    def __repr__(self) -> str:
        return "<Balancer>"
    
    @cached_property
    def versions(self) -> List[Union[BalancerV1, BalancerV2]]:
        return [v for v in [self.v1, self.v2] if v]

    @cached_property
    def v1(self) -> Optional[BalancerV1]:
        try: return BalancerV1()
        except ImportError: return None
    
    @cached_property
    def v2(self) -> Optional[BalancerV2]:
        try: return BalancerV2()
        except ImportError: return None

    #yLazyLogger(logger)
    def is_balancer_pool(self, token_address: AnyAddressType) -> bool:
        return await_awaitable(self.is_balancer_pool_async(token_address))

    #yLazyLogger(logger)
    async def is_balancer_pool_async(self, token_address: AnyAddressType) -> bool:
        return any(await asyncio.gather(*[v.is_pool_async(token_address) for v in self.versions]))
    
    #yLazyLogger(logger)
    async def get_pool_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        return await_awaitable(self.get_pool_price_async(token_address, block=block))

    #yLazyLogger(logger)
    async def get_pool_price_async(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        for v in self.versions:
            if await v.is_pool_async(token_address):
                return UsdPrice(await v.get_pool_price_async(token_address, block))

    #yLazyLogger(logger)
    @lru_cache(maxsize=None)
    def get_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        return await_awaitable(self.get_price_async(token_address, block=block))
    
    #yLazyLogger(logger)
    @alru_cache(maxsize=None)
    async def get_price_async(self, token_address: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        if await self.is_balancer_pool_async(token_address):
            return await self.get_pool_price_async(token_address, block=block)

        price = None
        
        if ( # NOTE: Only query v2 if block queried > v2 deploy block plus some extra blocks to build up liquidity
            (chain.id == Network.Mainnet and (not block or block > 12272146 + 100000))
            or (chain.id == Network.Fantom and (not block or block > 16896080))
            ): 
            price = await self.v2.get_token_price_async(token_address, block)
            if price:
                logger.debug(f"balancer v2 -> ${price}")
                return price

        if not price and chain.id == Network.Mainnet:      
            price = await self.v1.get_token_price_async(token_address, block)
            if price:
                logger.debug(f"balancer v1 -> ${price}")
                return price
        

balancer_multiplexer = BalancerMultiplexer()
